from unittest import TestCase

from xcover_lang.utils import eval_expr


class DictionaryTestCase(TestCase):

    def setUp(self):
        self.expr = '''
        foo = {
            "UK": 1,
            "CN": 2,
            "AU": 3
        }
        '''

    def test_dictionary(self):
        r, context = eval_expr(self.expr)

        self.assertDictEqual(context['foo'], {
            "UK": 1,
            "CN": 2,
            "AU": 3
        })

    def test_dictionary_lookup(self):
        expr = self.expr + '''
        au = foo["AU"]
        uk = foo["UK"]
        ru = foo["RU"]
        jp = foo["JP"] or 0
        '''
        r, context = eval_expr(expr)
        self.assertEqual(context['uk'], 1.0)
        self.assertEqual(context['au'], 3.0)
        self.assertEqual(context['jp'], 0.0)
        self.assertIsNone(context['ru'])



from unittest import TestCase

from xcover_lang.utils import eval_expr


class ArithmeticTestCase(TestCase):

    def test_additions(self):
        test_set = {
            '1 + 1': 2,
            '1 + 1 + 1': 3,
            '1 + 0 + 1': 2,
            '1.5 + 1': 2.5,
            '-1 + 2': 1
        }
        for expr, expected_result in test_set.items():
            result, _ = eval_expr(expr)
            self.assertEqual(result, expected_result)

    def test_additions_between_strings(self):
        test_set = {
            '"a" + "b"': 'ab',
            '"a" + "b" + "c"': 'abc'
        }
        for expr, expected_result in test_set.items():
            result, _ = eval_expr(expr)
            self.assertEqual(result, expected_result)

    def test_subtract(self):
        test_set = {
            '1 - 1': 0,
            '1 - 2': -1,
            '0 - 2': -2,
            '-1 - 2': -3
        }
        for expr, expected_result in test_set.items():
            result, _ = eval_expr(expr)
            self.assertEqual(result, expected_result)

    def test_multiply(self):
        test_set = {
            '1 * 1': 1,
            '1 * 0': 0,
            '1 * 1.33333': 1.33333,
            '-1 * 2': -2,
            '2 * -1': -2,
        }
        for expr, expected_result in test_set.items():
            result, _ = eval_expr(expr)
            self.assertEqual(result, expected_result)

    def test_div(self):
        test_set = {
            '1 / 1': 1,
            '3 / 2': 1.5,
            '0 / 1': 0,
            '2 / -1': -2,
            '-2 / 2': -1,
        }
        for expr, expected_result in test_set.items():
            result, _ = eval_expr(expr)
            self.assertEqual(result, expected_result)

    def test_parentheses_rules(self):
        test_set = {
            '1 * 2 + 1': 3,
            '3 * (1 + 2) ': 9,
            '(3 * 3) + (1 + 2)': 12,
        }
        for expr, expected_result in test_set.items():
            result, _ = eval_expr(expr)
            self.assertEqual(result, expected_result)

    def test_arithmetic_precedences(self):
        test_set = {
            '1 + 2 * 2': 5,
            '3 * (1 + 2)': 9,
            '3 * (1 + 2) * 2 + 2 * 2': 22,
        }
        for expr, expected_result in test_set.items():
            result, _ = eval_expr(expr)
            self.assertEqual(result, expected_result)

    def test_negative_numbers(self):
        self.assertEqual(-1, eval_expr('-1')[0])
        self.assertEqual(-0, eval_expr('-0')[0])


class LogicalOperatorTestCase(TestCase):

    def test_or_operator(self):
        self.assertEqual(eval_expr('1 or 2')[0], 1.0)
        self.assertEqual(eval_expr('0 or 0 or 1')[0], 1.0)

    def test_and_operator(self):
        self.assertEqual(eval_expr('1 and 2')[0], True)
        self.assertEqual(eval_expr('1 and 2 and 0')[0], False)


class EqualitiesTestCase(TestCase):

    def test_inline_equalities(self):
        self.assertEqual(eval_expr('1 == 1')[0], True)
        self.assertEqual(eval_expr('-1 == -1')[0], True)

        self.assertEqual(eval_expr('3 >= 3')[0], True)
        self.assertEqual(eval_expr('3 <= 3')[0], True)

    def test_inline_inequalities(self):
        self.assertEqual(eval_expr('3 <> 3')[0], False)
        self.assertEqual(eval_expr('3 <> 2')[0], True)

        self.assertEqual(eval_expr('1 > 0')[0], True)
        self.assertEqual(eval_expr('1 < 0')[0], False)



grammar = r"""
    start: instruction+
         | expression                       -> eval_expression    

    instruction: NAME "=" expression        -> assign_var

    expression: term
              | sum
              | or_test

    ?or_test: and_test
            | or_test "or" and_test         -> eval_or_expression 
    ?and_test: not_test 
             | and_test "and" not_test      -> eval_and_test
    
    ?not_test: comparison
             | "not" not_test               -> eval_not

    ?comparison: term "in" term             -> eval_membership_test
               | term _comp_op term         -> eval_equalities_test
               | term  
    
    term: object
        | list
        | string
        | "true"                             -> true
        | "false"                            -> false
        | "null"                             -> null    
        | "None"                             -> null
        | "True"                             -> true
        | "False"                            -> false
        | NAME"["term"]"                     -> access_object
        | atom
    
    list : "[" [term ("," term)*] "]"
    object:"{" [pair ("," pair)*] "}"
    pair : string ":" term
    string : ESCAPED_STRING

    ?sum: product
        | sum "+" product                    -> term_add
        | sum "-" product                    -> term_subtract
    
    ?product: atom
            | product "*" atom               -> term_mul
            | product "/" product            -> term_div
          
    ?atom: NUMBER                            -> number
         | "-" atom                          -> neg
         | "(" sum ")"
         | NAME "(" args ")"                 -> func_call
         | NAME                              -> access_var
         | string
         
    !_comp_op: "<"
             | ">"
             | "=="
             | ">="
             | "<="
             | "<>"       
     
    args: argvalue ("," argvalue)*
    argvalue: expression
            | NAME "=" expression           -> access_param


    %import common.CNAME -> NAME
    %import common.ESCAPED_STRING
    %import common.STRING_INNER
    %import common.WORD
    %import common.NUMBER
    
    %import common.SIGNED_NUMBER
    %import common.WS
    %ignore WS
"""



import operator
from datetime import datetime
from functools import reduce
from lark import (
    Transformer,
    Tree,
)
from lark.tree import Discard
import arrow
from arrow.parser import ParserError

def recrusive_get_num(tree):
    if not isinstance(tree, Tree):
        return tree
    val = recrusive_get_num(tree.children[0])
    if val:
        if str(val).isdigit():
            return float(val)
        else:
            return val
    else:
        return val


class TreeTransformer(Transformer):
    number = float

    def __init__(self, vars):
        self.vars = vars or {}

    def term_add(self, terms):
        return reduce(operator.__add__, (map(recrusive_get_num, terms)))

    def term_subtract(self, terms):
        return reduce(operator.__sub__, map(recrusive_get_num, terms))

    def term_mul(self, terms):
        return reduce(operator.mul, map(recrusive_get_num, terms), 1)

    def term_div(self, terms):
        return reduce(operator.truediv, map(recrusive_get_num, terms))

    def neg(self, term):
        return -recrusive_get_num(term[0])

    def string(self, s):
        return s[0][1:-1]

    def number(self, n):
        return float(n[0].value)

    def func_call(self, terms):
        func_name = terms[0].value
        args = list(map(recrusive_get_num, terms[1].children))
        if func_name == 'max':
            return max(args)
        elif func_name == 'min':
            return min(args)
        if func_name == 'count':
            return len([item for sublist in args for item in sublist])
        elif func_name == 'if_cond':
            return args[1] if args[0] else args[2]
        elif func_name == 'reduce_sum':
            return reduce(operator.__add__, (map(lambda x: x.get(args[1]), args[0])))
        elif func_name == 'number_of_days':
            dt_format = None
            rounding = None

            if not isinstance(args[0], str) or not isinstance(args[1], str):
                raise Exception("The first two params need to be start_date and end_date")

            for each in args:
                if isinstance(each, dict): # means we have some function parameters
                    for key, val in each.items() :
                        if key == 'format':
                            dt_format = val
                        elif key == 'round':
                            rounding = val
                        else:
                            raise Exception("Only format and round are supported as params")

            try:
                if dt_format is None:
                    date1 = arrow.get(args[0])
                    date2 = arrow.get(args[1])
                else:
                    date1 = arrow.get(args[0], dt_format)
                    date2 = arrow.get(args[1], dt_format)
            except ParserError as msg:
                raise Exception(msg)
            
            # calculate number of days between these two dates
            td = abs(date1-date2)
            days = abs(date1-date2).days
            # calculate number of hours
            hours = td.seconds/3600
            if rounding is not None and rounding == 'up' and hours >= 12:
                days += 1
            if rounding is not None and rounding == 'down' and hours <= 12:
                days -= 1

            return days
        else:
            raise Exception('Unkown function {0}'.format(func_name))

    def eval_equalities_test(self, terms):
        left = recrusive_get_num(terms[0])
        right = recrusive_get_num(terms[2])

        op = terms[1].value

        ops_lookup = {
            '<': lambda a, b: a < b,
            '>': lambda a, b: a > b,
            '==': lambda a, b: a == b,
            '>=': lambda a, b: a >= b,
            '<=': lambda a, b: a <= b,
            '<>': lambda a, b: a != b,
        }
        comparator = ops_lookup.get(op, None)
        if not comparator:
            raise Exception('Unknown operator, supports only < > == >= <= <>')

        return comparator(left, right)

    def assign_var(self, _vars):
        if isinstance(_vars[1], Tree):
            items = []
            for c in _vars[1].children:
                try:
                    items.append(self.transform(c) if isinstance(c, Tree) else c)
                except Discard:
                    pass

            self.vars[_vars[0].value] = items[0]
            return items[0]
        else:
            self.vars[_vars[0].value] = _vars[1]
            return _vars[1]
    
    def access_param(self, _vars):
        return {_vars[0].value: _vars[1].children[0]}

    def bool_eval(self, terms):
        return terms[0] or terms[1]

    def access_object(self, _vars):
        return self.vars.get(_vars[0], {}).get(str(_vars[1]))

    def access_var(self, name):
        return self.vars[name[0].value]

    def term(self, term):
        return term[0]

    def eval_expression(self, tree):
        return tree[0].children[0]

    def eval_or_expression(self, terms):

        if len(terms) == 1:
            result = True
            for term in terms:
                val = recrusive_get_num(term)
                if not bool(val):
                    result = False
                    break
        else:
            result = None
            for term in terms:
                val = recrusive_get_num(term)
                if bool(val):
                    return val
                else:
                    result = val
        return result

    def eval_and_test(self, terms):
        result = True
        for term in terms:
            val = recrusive_get_num(term)
            if not bool(val):
                result = False
                break
        return result

    def eval_not(self, terms):
        return not recrusive_get_num(terms[0])

    def eval_membership_test(self, terms):
        return terms[0] in terms[1]

    list = list
    pair = tuple
    object = dict

    null = lambda self, _: None
    true = lambda self, _: True
    false = lambda self, _: False



from lark import Lark
from xcover_lang.transformers import TreeTransformer
from xcover_lang.grammar import grammar


def eval_expr(expr, context=None):
    context = context or {}
    lark_parser = Lark(grammar, parser='earley', start='start', lexer='dynamic')
    tree = lark_parser.parse(expr)
    transformer = TreeTransformer(context)
    result = transformer.transform(tree)

    return result, transformer.vars



from unittest import TestCase

from xcover_lang.utils import eval_expr


class VariableTestCase(TestCase):

    def test_assignment(self):
        expr = '''
        a = 1
        b = "string"
        c = {"foo": "bar"}
        d = [1, 2, 3]
        '''
        r, context = eval_expr(expr)

        self.assertEqual(context['a'], 1)
        self.assertEqual(context['b'], 'string')
        self.assertDictEqual(context['c'], {"foo": 'bar'})


class BuiltInFunctionTestCase(TestCase):
    def test_min(self):
        result, _ = eval_expr('min(2, 0)')
        self.assertEqual(result, 0)

    def test_max(self):
        result, _ = eval_expr('max(2, 0)')
        self.assertEqual(result, 2)

    def test_count(self):
        result, _ = eval_expr('count([1, 2, 3])')
        self.assertEqual(result, 3)

    def test_membership_testing(self):
        result, _ = eval_expr('"UK" in ["UK", "US"]')
        self.assertTrue(result)
        result, _ = eval_expr('"UK" in ["US"]')
        self.assertFalse(result)

    def test_if_cond_inline_test(self):
        self.assertEqual(eval_expr(
            '''
            if_cond(1 == 1, "yes", "no")
            '''
        )[0], "yes")

    def test_multiple_if_conditions(self):
        context = eval_expr(
            '''
            variable1 = if_cond(1 == 1, "yes", "no")
            variable2 = if_cond(variable1 == "yes", "yes2", "no2")
            '''
        )[1]
        expected = {'variable2': 'yes2'}
        self.assertTrue(set(expected.items()).issubset(set(context.items())))

    def test_if_cond_code_block_test(self):
        self.assertEqual(eval_expr(
            '''
            a = 1
            b = 2
            result = if_cond(a == b, "yes", "no")
            '''
        )[1]['result'], "no")

        # Test "and" in expression and False assignment
        # Return a number instead of a string
        self.assertEqual(eval_expr(
            '''
            a = 1
            b = False
            result = if_cond(a and b, "yes", 1)
            '''
        )[1]['result'], 1)

    def test_reduce_sum(self):
        """
        reduce_sum computes the sum of an attribute over a list of objects
        """
        result, context = eval_expr(
            '''
            dic = [{"value": 1}, {"value": 2}]
            result = reduce_sum(dic, "value")
            '''
        )
        self.assertEqual(3, context['result'])

    def test_number_of_days_without_params(self):
        """
        compares the number of days between two dates
        """
        result, context = eval_expr('''result = number_of_days("2017-01-10", "2017-01-20")''')
        self.assertEqual(10, context['result'])

    def test_number_of_days_with_only_rounding_params(self):
        """
        compares the number of days between two dates
        """
        result, context = eval_expr('''result = number_of_days("2017-01-10 11:00:00", "2017-01-11 10:00:00", round = "up")''')
        self.assertEqual(1, context['result'])

    def test_number_of_days_with_params(self):
        """
        compares the number of days between two dates
        """
        result, context = eval_expr('''result = number_of_days("2017-01-10", "2017-01-20", format = "YYYY-MM-DD", round = "up")''')
        self.assertEqual(10, context['result'])

    def test_number_of_days_rounding_up(self):
        result, context = eval_expr('''result = number_of_days("2017-01-10 11:00:00", "2017-01-11 10:00:00", format = "YYYY-MM-DD HH:mm:ss", round = "up")''')
        self.assertEqual(1, context['result'])

    def test_number_of_days_rounding_down(self):
        result, context = eval_expr('''result = number_of_days("2017-01-10 11:00:00", "2017-01-11 10:00:00", format = "YYYY-MM-DD HH:mm:ss", round = "down")''')
        self.assertEqual(0, context['result'])





# coding: utf-8

"""
    XCover Expression Language
"""

from setuptools import setup, find_packages  # noqa: H301

NAME = "xcover_expression_language"
VERSION = "0.0.6"
# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = [
    "lark-parser==0.5.6",
    "arrow==0.12.1"
]

setup(
    name=NAME,
    version=VERSION,
    description="A DSL that evaluates against Python API",
    author_email="",
    url="",
    keywords=["expression", "string expression", "XCover"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description="""\
    A DSL that evaluates against Python API.
    """
)
